# -*- coding: utf-8 -*-
"""
Created on Sun Mar  8 15:38:44 2020

@author: Wolfgang Dvorak
"""
import queue
from typing import List
from typing import Tuple

import igraph
import pulp

print("Dung mod for igraph 0.7.x")


def _extract(raw_string: str, start_marker: int, end_marker: int) -> str:
    start = raw_string.index(start_marker) + len(start_marker)
    end = raw_string.index(end_marker, start)
    return raw_string[start:end]

def read_apx(filename: str) -> igraph.Graph:
    """Reads an AF from a File in apx format"""
#    eol="\n"6
    comment_char = "#"
    sep = "."
    if isinstance(filename, str):
        file = open(filename, "r")
        framework = igraph.Graph(directed=1)
        # We first create a list of all edges
        edges = []
        i = 0
        for line in file:
            i += 1
            try:
                for atom in line.split(sep):
                    atom = atom.strip()
                    if len(atom) == 0:
                        continue
                    if atom.startswith(comment_char):
                        break
                    if atom.startswith("arg("):
                        framework.add_vertex(name=_extract(atom, "(", ")"))
                    elif atom.startswith("att("):
                        edge = _extract(atom, "(", ")")
                        edge = edge.split(",")
                        # Adding edges directlyy to the graph is too slow
                        #af.add_edge(str(edge[0].strip()), str(edge[1].strip()))
                        edges += [(str(edge[0].strip()), str(edge[1].strip()))]
                    else: print("Line", str(i)+":", "Invalid atom:", atom)
            except:
                print("Parse error in line: ", i) 
            #if (i % 1000 == 0): print("Line", str(i))
        file.close()
        #Finally add the edges to the graph
        framework.add_edges(edges)
    return framework

def compute_grounded_ext(framework: igraph.Graph, inplace: int = 0) -> List[str]:
    """Computes the grounded extension of an AF"""
    grounded_ext = []

    # Queue that contains unprocessed defended arguments
    unprocessed_in = queue.Queue(maxsize=0)

    if inplace == 0:
        framework = framework.copy()

    # Start with the unattacked arguments
    for arg in framework.vs.select(_indegree=0):
        unprocessed_in.put(arg)

    # Iteratively process defended arguments
    while not unprocessed_in.empty():
        argument = unprocessed_in.get()
#        print("Process vertex:", argument["name"])
        grounded_ext += [argument["name"]]
        argument["status"] = "in"
#        print("Grounded:", grounded_ext)
        for att in framework.incident(argument,mode="OUT"):
            out_vertex = framework.vs[framework.es[att].target]
            if out_vertex["status"] != "out":
                out_vertex["status"] = "out"
                for att2 in framework.incident(out_vertex,mode="OUT"):
                    undec_vertex = framework.vs[framework.es[att2].target]
                    if undec_vertex["status"]:
                        if undec_vertex["status"] != "out":
                            undec_vertex["status"] -= 1
                    else:
                        undec_vertex["status"] = undec_vertex.indegree()-1
                    if undec_vertex["status"] == 0:
                        unprocessed_in.put(undec_vertex)
    return grounded_ext


def cone_of_influence(framework: igraph.Graph,
                      target: str = "target",
                      inplace: int = 0) -> igraph.Graph:
    """Deletes all arguments that can not reach the given target argument"""
    cone_arg = []
    con_att = []
    unprocessed_vertex = queue.Queue(maxsize=0)

    if inplace == 0:
        framework = framework.copy()

    for arg in framework.vs.select(name_eq=target):
        unprocessed_vertex.put(arg)
        cone_arg += [arg["name"]]
        arg["status"] = "visited"

    while not unprocessed_vertex.empty():
        arg = unprocessed_vertex.get()

        for att in framework.incident(arg, mode="IN"):
            in_arg = framework.vs[framework.es[att].source]
            con_att += [(in_arg["name"], arg["name"])]
            if in_arg["status"] != "visited":
                in_arg["status"] = "visited"
                cone_arg += [in_arg["name"]]
                unprocessed_vertex.put(in_arg)

    cone = igraph.Graph(directed=1)
    cone.add_vertices(cone_arg)
    cone.add_edges(con_att)
    del framework.vs["status"]
    return cone

def compute_min_strongly_admissible_ext(framework: igraph.Graph,
                                        target: str = "target",
                                        inplace: int = 0,
                                        encoding: int = 1,
                                        opt_weight: Tuple[int, int] = (1, 0)
                                        )-> Tuple[List[str], List[str]]:
    """Computes a minimum size strongly admissible set containing the given target argument"""

    # We first compute A grounded extension and the corresponding sets for the ILP solver

    # arguments in the grounded extension
    grounded_ext_in = []
    # unattacked arguments
    unattacked_arg = []
    # arguments that attack the grounded extension
    grounded_ext_att = []
    # attacks attacking the grounded extension
    grounded_ext_in_attacks = []
    # attacks defending the grounded extension
    grounded_ext_out_attacks = dict()

    # Queue that contains unprocessed defended arguments
    unprocessed_in = queue.Queue(maxsize=0)

    # Reduce to the relevant part of the AF
    framework = cone_of_influence(framework, target, inplace)

    # Initiate the "write" dictionary
    framework.vs["write"] = ["dummy"]

    # Start with the unattacked arguments
    for arg in framework.vs.select(_indegree=0):
        unprocessed_in.put(arg)
        unattacked_arg += [arg["name"]]

     # Iteratively process defended arguments
    while not unprocessed_in.empty():
        argument = unprocessed_in.get()
#        print("Process vertex:", argument["name"])
        grounded_ext_in += [argument["name"]]
        argument["status"] = "in"
#        print("Grounded:", grounded_ext)
        # update the attackers of the grounded ext
        for att in framework.incident(argument, mode="IN"):
            in_vertex = framework.vs[framework.es[att].source]
            if in_vertex["write"] != "done":
                grounded_ext_att += [in_vertex["name"]]
                in_vertex["write"] = "done"
            grounded_ext_in_attacks += [(in_vertex["name"], argument["name"])]

        # continue with the computation of the grounded extension
        for att in framework.incident(argument, mode="OUT"):
            out_vertex = framework.vs[framework.es[att].target]
            if out_vertex["status"] != "out":
                out_vertex["status"] = "out"
                for att2 in framework.incident(out_vertex, mode="OUT"):
                    undec_vertex = framework.vs[framework.es[att2].target]
                    if undec_vertex["status"]:
                        if undec_vertex["status"] != "out":
                            undec_vertex["status"] -= 1
                    else:
                        undec_vertex["status"] = undec_vertex.indegree()-1
                    if undec_vertex["status"] == 0:
                        unprocessed_in.put(undec_vertex)
            # update the defenders of the grounded ext
            defenders = []
            for att2 in framework.incident(out_vertex, mode="IN"):
                in_vertex = framework.vs[framework.es[att2].source]
                if in_vertex["status"] == "in":
                    defenders += [in_vertex["name"]]
            if defenders:
                grounded_ext_out_attacks[out_vertex["name"]] = defenders
    del unprocessed_in

    if inplace != 0:
        del framework
    elif len(unattacked_arg) != 0:
        del framework.vs["status"]
        del framework.vs["write"]

    if target not in grounded_ext_in:
#        print(grounded_ext_in)
        return "UNSAT", "UNSAT"

    # Start with constructing the ILP
    #TOL = 0.0001
    TOL=0.0001
    
#    print("In Args:", grounded_ext_in)
#    # unattacked arguments
#    print("Unattacked Args:", unattacked_arg)
#    # arguments that attack the grounded extension
#    print("Attacking Args:", grounded_ext_att)
#    # attacks attacking the grounded extension
#    print("Attacks:", grounded_ext_in_attacks)
#    # attacks defending the grounded extension
#    print("Attacks:", grounded_ext_out_attacks)

    # Minimum over the upper bounds for the number of iterations of the characteristic function
    k = min(len(grounded_ext_in), len(grounded_ext_att)+1)

    print("Best upper bound for the number of iterations: ", k)

    linear_program = pulp.LpProblem("StrongAdmissible", pulp.LpMinimize)

    if encoding == 1:
        pro_args = pulp.LpVariable.dicts("ProArgs", [(i, j) for i in grounded_ext_in for j in range(k)], 0, 1, pulp.LpBinary)
        con_args = pulp.LpVariable.dicts("ConArgs", grounded_ext_att, 0, 1, pulp.LpBinary)

        print("Variables defined.")

        # Objective function
        linear_program += opt_weight[0]*pulp.lpSum(pro_args[i, k-1] for i in grounded_ext_in) + opt_weight[1]*pulp.lpSum(con_args[i] for i in grounded_ext_att)

        print("Objective stated.")

        # target argument
        linear_program += pro_args[target, k-1] == 1

        print("target constraint added.")

        # Arguments in the extension are defended
        for j in range(1, k):
            for (out_arg, in_arg) in grounded_ext_in_attacks:
                linear_program += pro_args[in_arg, j] <= pulp.lpSum(pro_args[x, j-1] for x in grounded_ext_out_attacks[out_arg])

        for in_arg in grounded_ext_in:
            if in_arg not in unattacked_arg:
                linear_program += pro_args[in_arg, 0] <= TOL

        print("In args are defended.")

        # out arguments are attacked
        for (out_arg, in_arg) in grounded_ext_in_attacks:
            linear_program += pro_args[in_arg, k-1] <= con_args[out_arg]

        print("Out args are attacked.")

        # The characteristic function is monotone
        for j in range(k-1):
            for i in  grounded_ext_in:
                linear_program += pro_args[i, j] <= pro_args[i, j+1]

        print("monotonicity ensured.")

        print("LP constructed")

        # start solver
        solver = pulp.CPLEX_CMD(path="/files/00-DBAI/sw/IBM-Cplex/cplex/bin/x86-64_linux/cplex")
        linear_program.solve(solver)
        #linear_program.solve()

        print("LP solved")

        # Convert solution to Extension
        str_adm_ext = []
        str_adm_att = []
        for i in  grounded_ext_in:
            try:
                if pro_args[i, k-1].varValue > 0.99:
                    str_adm_ext += [i]
            except Exception as e: print(e)
            
#        for j in range(k):
#            strAdmExt=[]    
#            for i in grounded_ext_in:    
#                if pro_args[i,j].varValue > 0.99:
#                    strAdmExt+=[i]
#            print("Str. Adm. Level in ", j, ":", strAdmExt)  

    else:
        pro_args = pulp.LpVariable.dicts("ProArgs", [(i, j) for i in grounded_ext_in for j in range(k)], 0, 1, pulp.LpBinary)
        con_args = pulp.LpVariable.dicts("ConArgs", [(i, j) for i in grounded_ext_att for j in range(k-1)], 0, 1, pulp.LpBinary)

        print("Variables defined.")

        # Objective function
        linear_program += opt_weight[0]*pulp.lpSum(pro_args[i, k-1] for i in grounded_ext_in) + opt_weight[1]*pulp.lpSum(con_args[i, k-2] for i in grounded_ext_att)

        print("Objective stated.")

        # target argument
        linear_program += pro_args[target, k-1] == 1

        print("target constraint added.")

        # Arguments in the extension are defended
        for j in range(1, k):
            for (out_arg, in_arg) in grounded_ext_in_attacks:
                linear_program += pro_args[in_arg, j] <= con_args[out_arg, j-1]

        #for (s, t) in grounded_ext_in_attacks:
                #linear_program += pro_args[t, 0] == 0

        for in_arg in grounded_ext_in:
            if in_arg not in unattacked_arg:
                linear_program += pro_args[in_arg, 0] <= TOL

        print("In args are defended.")

        # out arguments are attacked
        for j in range(k-1):
            for out_arg in grounded_ext_att:
                linear_program += con_args[out_arg, j] <= pulp.lpSum(pro_args[s, j] for s in grounded_ext_out_attacks[out_arg])

        print("Out args are attacked.")

        # The characteristic function is monotone
        for j in range(k-1):
            for i in  grounded_ext_in:
                linear_program += pro_args[i, j] <= pro_args[i, j+1]
        for j in range(k-2):
            for i in grounded_ext_att:
                linear_program += con_args[i, j] <= con_args[i, j+1]

        print("monotonicity ensured.")

        print("LP constructed")

        # start solver
        solver = pulp.CPLEX_CMD(path="/files/00-DBAI/sw/IBM-Cplex/cplex/bin/x86-64_linux/cplex")
        linear_program.solve(solver)
        #linear_program.solve()

        print("LP solved")

        # Convert solution to Extension
        str_adm_ext = []
        str_adm_att = []
        for i in  grounded_ext_in:
            try:
                if pro_args[i, k-1].varValue > 0.99:
                    str_adm_ext += [i]
            except Exception as e: print(e)
    return str_adm_ext, str_adm_att
